/*
                  ___  _____________
                 / _ \/ __/  _/ ___/
                / , _/ _/_/ // /__
               /_/|_/_/ /___/\___/

			RedKod File Integrity Checker

				   Version 0.5
			  Compatible Win32/*Nux

                  RedKod Team
              http://www.redkod.com/


Coder: R-e-D
Mail : r-e-d@redkod.com

*/

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#include <windows.h>
#include <fcntl.h>
#include <io.h>
#include <time.h>
#else
#include <unistd.h>
#include <sys/time.h>
#include <sys/stat.h>
#endif

/* md5 header */
#include "md5.h"

#define CONFIG_FILE		"rfic.conf"
#define MD5HASH_SAVE	"rfic.save"

#ifdef _WIN32
#define SEPARATOR	'\\'
#else
#define SEPARATOR	'//'
#endif

#define CONFIGFILE_ARG	1
#define LINEBUF_SIZE	256

/* strlcpy */
size_t strlcpy(char *dst, const char *src, size_t size);

/* Output */
static FILE *out;

struct rfic_config {
    char **files;		// path of all files
    char **md5sum;		// MD5 Sum
    long *file_size;		// size of file
    unsigned int nb_of_files;	// Number of files in the configuration files
    unsigned int nb_of_sums;

} rfic_config;


/* Generate a MD5 Sum */
unsigned char *MD5Sum(const char *file)
{
    unsigned char buffer, *md5sum = NULL;
    unsigned int i;
    FILE *fp = NULL;
    struct MD5Context md5c;

    MD5Init(&md5c);

    if ((fp = fopen(file, "rb")) == NULL) {
	fprintf(out, "Error while opening %s.\n", file);
	return (NULL);
    }
#ifdef _WIN32
    /* I/O For Win32 */
    _setmode(_fileno(fp), _O_BINARY);
#endif

    md5sum = (unsigned char *) malloc(16 * sizeof(unsigned char));

    while ((i = fread(&buffer, 1, sizeof(unsigned char), fp)) > 0)
	MD5Update(&md5c, &buffer, (unsigned) i);

    MD5Final(md5sum, &md5c);

    fclose(fp);

    return (md5sum);
}

char *get_and_save_md5sum(const unsigned char *md5sum)
{
    unsigned char *buf = NULL;
    unsigned int i;

    buf = (char *) malloc(256 * sizeof(char));
    memset(buf, 0x0, 256 * sizeof(char));

    for (i = 0; i < 16 * sizeof(unsigned char); i++) {
	sprintf(buf, "%s%02X", buf, md5sum[i]);
    }

    return (buf);
}

/* Display a MD5 Sum */
void DisplayMD5Sum(const unsigned char *md5sum)
{
    unsigned int i;

    fprintf(out, "{");
    for (i = 0; i < 16 * sizeof(unsigned char); i++)
	fprintf(out, "%02X", md5sum[i]);
    fprintf(out, "}");
}

/* Get the number of files in the configuration file */

int get_num_of_files(const char *file)
{
    FILE *fp = NULL;
    int value;
    unsigned int count = 0;

    fp = fopen(file, "rb");
    if (fp == NULL) {
	fprintf(out, "Error while loading configuration file : %s.\n",
		file);
	exit(-1);
    }

    while (!feof(fp)) {
	value = fgetc(fp);
	switch (value) {
	case '[':
	    count++;
	    break;
	case '{':
	    count++;
	    break;
	}
    }

    if (count < 1) {
	fprintf(out, "Bad params in configuration file.\n");
	exit(-1);
    }

    fclose(fp);

    return (count);
}


/* get all names of files int the configuration file */
struct rfic_config *parse_file(const char *parsed_file,
			       struct rfic_config *config)
{
#ifdef _WIN32
    FILE *tmpfp = NULL;
#else
    struct stat *status = NULL;
#endif

    FILE *fp = NULL;
    char line[LINEBUF_SIZE + 2];
    unsigned int i = 0, count = 0, len;

    if (!strcmp(parsed_file, (char *) MD5HASH_SAVE)) {
	config->nb_of_sums = get_num_of_files(parsed_file);
	config->md5sum = malloc((config->nb_of_sums) * 4);
    } else {
	config->nb_of_files =
	    get_num_of_files(parsed_file) - CONFIGFILE_ARG;
	config->files = malloc((config->nb_of_files + CONFIGFILE_ARG) * 4);
	config->file_size = malloc(config->nb_of_files * 4);
    }


    fp = fopen(parsed_file, "rb");
    if (fp == NULL) {
	fprintf(out, "Error while loading %s.\n", parsed_file);
	exit(-1);
    }

    while (!feof(fp)) {
	fgets(line, sizeof(line), fp);
	len = strlen(line);

	/* Overflow Control */
	if ((len >= 0) && (len <= LINEBUF_SIZE)) {
	    i++;
	    switch (*line) {
	    case '[':

		config->files[count] = (char *) malloc(256 * sizeof(char));
		memset(config->files[count], 0x0, 256 * sizeof(char));
		/* Get path of the file */
		sscanf(line, "[%[^]]]", config->files[count]);
		/* open the file */
		if (count >= CONFIGFILE_ARG) {
#ifdef _WIN32
		    tmpfp = fopen(config->files[count], "r");
		    if (tmpfp == NULL) {
			fprintf(stderr, "Error while opening \" %s \"\n",
				config->files[count]);
			free(config);
			fclose(tmpfp);
			exit(-1);
		    }
		    /* Get the size of the file */
		    config->file_size[count] = filelength(tmpfp->_file);
		    /* Close the file */
		    fclose(tmpfp);
#else
		    status = (struct stat *) malloc(sizeof(struct stat));
		    memset(status, 0x0, sizeof(struct stat));
		    stat(config->files[count], status);
		    config->file_size[count] = status->st_size;
		    free(status);
#endif
		}
		count++;
		break;

	    case '{':
		config->md5sum[count] =
		    (char *) malloc(256 * sizeof(char));
		memset(config->md5sum[count], 0x0, 256 * sizeof(char));
		sscanf(line, "{%[^]}]", config->md5sum[count]);
		count++;
		break;
	    case '#':		/* ignore comment line */
		continue;
	    default:
		printf("Parse Error: [Line %d]: %s \n", i, line);
		exit(-1);
		break;

		fseek(fp, strlen(line), strlen(line));
	    }
	} else {		/* Overflow @#! */

	    fprintf(stderr,
		    "Error: Buffer overflow while reading configuration file.\n");
	    fclose(fp);
	    free(config);
	    exit(-1);
	}
    }

    // fclose(fp); // Fucking bug

    return (config);
}

char *GetFileName(char *path)
{
    unsigned int i;
    char *ext = NULL;

    if (strchr(path, (char) SEPARATOR) != NULL) {
	for (i = strlen(path) - 1; i >= 0 && path[i] != (char) SEPARATOR;
	     i--)

	    ext =
		malloc((strlen(path) - CONFIGFILE_ARG + i) * sizeof(char));
	strcpy(ext, path + CONFIGFILE_ARG + i);
    } else
	return (path);

    return (ext);
}

int PortableCopyFile(const char *srcpath, const char *destpath)
{
    FILE *srcfile = NULL, *dstfile = NULL;
    int value;

    srcfile = fopen(srcpath, "rb");
    if (srcfile == NULL) {
	fprintf(stderr, "Error while opening source file : %s .\n",
		srcpath);
	return (-1);
    }

    dstfile = fopen(destpath, "wb");
    if (srcfile == NULL) {
	fprintf(stderr, "Error while opening destination file : %s .\n",
		destpath);
	fclose(srcfile);
	return (-1);
    }

    while ((value = fgetc(srcfile)) != EOF) {
	if (fputc(value, dstfile) == EOF)
	    break;
    }

    fclose(srcfile);
    fclose(dstfile);

    return (0);
}

int
RestoreFile(struct rfic_config *config, unsigned int corrupt_file_index)
{
    char **backuppath = NULL, *filename = NULL, *destpath = NULL;
    unsigned int len;

    filename = GetFileName((char *) config->files[corrupt_file_index]);

    len = strlen(config->files[0]) + strlen(filename) + 2;
    destpath = (char *) malloc(len);
    memset(destpath, 0x0, len);

    sprintf(destpath, "%s%s", config->files[0], filename);

    PortableCopyFile(destpath, config->files[corrupt_file_index]);

    fprintf(out, "The file \" %s \" has been restored !\n\n",
	    config->files[corrupt_file_index]);

    free(filename);
    free(destpath);

    return (0);
}


int BackupFiles(const char *configfiles, struct rfic_config *config)
{
    FILE *fp = NULL;
    unsigned int i;
    int value;
    char *dstfilepath = NULL, *filename = NULL;

    for (i = 0; i < config->nb_of_files; i++) {
	filename = GetFileName(config->files[CONFIGFILE_ARG + i]);

	fp = fopen(config->files[CONFIGFILE_ARG + i], "rb");
	if (fp == NULL) {
	    fprintf(stderr, "Error while opening \" %s \" .\n",
		    config->files[CONFIGFILE_ARG + i]);
	    fprintf(out, "Please, check your configuration file.\n");
	    free(filename);
	    free(config);
	    exit(-1);
	}

	value =
	    strlen(config->files[0]) + strlen(filename) + 2 * sizeof(char);

	/* Create the new path */
	dstfilepath = (char *) malloc(value);
	memset(dstfilepath, 0x0, value);
	sprintf(dstfilepath, "%s%s", config->files[0], filename);
	fprintf(out, "Copie de %s vers %s ...",
		config->files[CONFIGFILE_ARG + i], dstfilepath);

	/* Copy the file */
	PortableCopyFile(config->files[CONFIGFILE_ARG + i], dstfilepath);

	/* Display size */
	fprintf(out, "\n\tSize\t%d Octets\t[OK]\n",
		config->file_size[CONFIGFILE_ARG + i]);

	/* free the world */
	free(dstfilepath);
	free(filename);
	dstfilepath = NULL;
	filename = NULL;
    }

    return (0);
}

int main(int argc, char **argv)
{
    struct rfic_config *config = NULL;
    unsigned char *md5sum = NULL;
    char restore;
    unsigned int i, flag = 0;
    time_t t;

    out = stdout;

    /* Params */
    if (argc < 2) {
	fprintf(out, "Usage : rfic [options]\n");
	fprintf(out,
		"	-g	Generate all md5 sums and save to the md5 sums save file.\n");
	fprintf(out,
		"	-b	Backup all files present in the config files.\n");
	fprintf(out, "	-c	Check all md5 sums.\n");

	exit(-1);
    }

    /* Initialize time_t structure */
    time(&t);

    /* Loading Configuration File */
    config = (struct rfic_config *) malloc(sizeof(struct rfic_config));
    memset(config, 0x0, sizeof(struct rfic_config));

    fprintf(stdout, "Loading Configuration File : %s...\n",
	    (char *) CONFIG_FILE);
    config = parse_file((char *) CONFIG_FILE, config);

    switch (argv[1][1]) {

    case 'b':			/* Backup all files */

#ifdef _WIN32
	CreateDirectory(config->files[0], NULL);
#else
	mkdir(config->files[0], 0777);
#endif
	fprintf(out, "Backup all files in %s .\n", config->files[0]);
	BackupFiles((char *) CONFIG_FILE, config);

	break;

    case 'g':			/* Generate all MD5 sums and save to MD5HASH_SAVE */

	/* open MD5HASH_SAVE */
	out = fopen((char *) MD5HASH_SAVE, "w");
	if (out == NULL) {
	    fprintf(stderr, "Error while opening %s.\n",
		    (char *) MD5HASH_SAVE);
	    free(config);
	    exit(-1);
	}

	/* Prelude */
	fprintf(out,
		"#######################################\n# RFIC By R-e-D (r-e-d@redkod.com)\n# MD5 HASH SAVE FILE\n");
	fprintf(out,
		"# Generated at %s# Do not edit this file !\n#######################################\n",
		ctime(&t));

	/* Generate all MD5 sums */
	fprintf(stdout, "Generating All MD5 Sum and save in %s...\n\n",
		(char *) MD5HASH_SAVE);

	for (i = 0; i < config->nb_of_files; i++) {
	    md5sum = MD5Sum(config->files[CONFIGFILE_ARG + i]);

	    if (md5sum != NULL) {
		DisplayMD5Sum(md5sum);
		fprintf(out, "\t%s\n", config->files[CONFIGFILE_ARG + i]);
	    }
	    fprintf(stdout, ".");
	}

	out = stdout;
	fprintf(out, " [OK]\n");
	break;

    case 'c':			/* Check all MD5 sums */

	/* Loading all saved md5 sums */
	fprintf(out, "Loading Saved MD5 Hash File : %s...\n",
		(char *) MD5HASH_SAVE);
	parse_file((char *) MD5HASH_SAVE, config);

	/* Check if all ok */
	if (config->nb_of_files != config->nb_of_sums) {
	    fprintf(stderr,
		    "Error, Number of files != Number of MD5 sums.\n");
	    free(config);
	    exit(-1);
	}

	/* Check all sums now */
	fprintf(out, "Global Checking for all MD5 sums...\n\n");

	for (i = 0; i < config->nb_of_files; i++) {
	    md5sum = MD5Sum(config->files[CONFIGFILE_ARG + i]);

	    if (md5sum != NULL) {
		if (!strcmp
		    (config->md5sum[i], get_and_save_md5sum(md5sum)))
		    fprintf(out, "[OK] %d: %s\t%s\n", CONFIGFILE_ARG + i,
			    get_and_save_md5sum(md5sum),
			    config->files[CONFIGFILE_ARG + i]);
		else {
		    fprintf(out,
			    "\n[WARNING!] %d: %s\n\tHour\t\t%s\tSaved MD5 Sum\t%s\n\tNew MD5 Sum\t%s\n\n",
			    CONFIGFILE_ARG + i,
			    config->files[CONFIGFILE_ARG + i], ctime(&t),
			    config->md5sum[i],
			    get_and_save_md5sum(md5sum));

		    fprintf(out, "Restore the original file ? (Y/N)\n");

		    scanf("%c", &restore);

		    if (restore == 'Y') {
			fprintf(out,
				"Ok ! Now Restoring the original file with backup.\n");
			RestoreFile(config, CONFIGFILE_ARG + i);
		    } else
			fprintf(out,
				"The file \" %s \" has not been restored.\n\n",
				config->files[CONFIGFILE_ARG + i]);
		}
	    }
	}
	/* End */
	break;
    }

    /* Free */
    free(config);

    return (0);
}
